%  Copyright (C) 2002-2004 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; if not, write to the Free Software Foundation,
%  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
\subsection{darcs reportbug}
\begin{code}
module ReportBug ( reportbug ) where
import IO ( stdout, hFlush )
import Monad ( liftM )
import Maybe ( catMaybes )

import DarcsCommands ( DarcsCommand(..), nodefaults )
import DarcsArguments ( DarcsFlag( Output, Target ),
                        get_cc, output, cc, target,
                        from_opt, get_author, author, sign,
                        edit_file,
                      )
import Patch ( namepatch, hunk, adddir, addfile, join_patches, adddeps )
import SlurpDirectory ( empty_slurpy )
import External ( signString, sendEmail, )
import FastPackedString ( unpackPS, readFilePS, linesPS, unlinesPS, packString )
import Lock ( writeBinFile, withTemp )
import PatchBundle ( make_bundle, scan_context )
import Context ( context )
import IsoDate ( getIsoDateTime )
import SelectChanges ( promptChar )
#include "impossible.h"
\end{code}
\begin{code}
reportbug_description :: String
reportbug_description =
 "Report a bug in darcs.\n"
\end{code}

\options{reportbug}

\haskell{reportbug_help}
\begin{code}
reportbug_help :: String
reportbug_help =
 "Reportbug is used to send a bug report regarding darcs.  By default, the\n"++
 "bug report is sent by email, although you may save it to a file.\n"
\end{code}
\begin{code}
reportbug :: DarcsCommand
reportbug = DarcsCommand {command_name = "reportbug",
                          command_help = reportbug_help,
                          command_description = reportbug_description,
                          command_extra_args = 0,
                          command_extra_arg_help = [],
                          command_command = reportbug_cmd,
                          command_prereq = \_ -> return $ (Just "", ""),
                          command_get_arg_possibilities = return [],
                          command_argdefaults = nodefaults,
                          command_darcsoptions = [from_opt, author,
                                                  target,cc,output,sign]}
\end{code}
\begin{code}
reportbug_cmd :: [DarcsFlag] -> [String] -> IO ()
reportbug_cmd input_opts [] = do
  opts <- decide_on_behavior input_opts
  wantfile <- wants_output opts
  report <- withTemp $ \f -> do writeBinFile f "SHORT BUG NAME HERE\n\n"
                                appendFile f "LONGER DESCRIPTION"
                                edit_file f
                                linesPS `liftM` readFilePS f
  from <- get_author opts
  date <- getIsoDateTime
  let patchname = "Bug report: " ++ unpackPS (head report) ++
                  " (" ++ from ++ " - " ++ date ++ ")"
      context_infos = map fst $ concat $ scan_context $ packString context
      thepatch = (namepatch date patchname from [] $ join_patches
                  [adddir $ "./bugs/"++date,
                   addfile $ "./bugs/"++date++"/"++date,
                   hunk ("./bugs/"++date++"/"++date) 1
                   (return []) (return report)]) `adddeps` context_infos
  bundle <- signString opts $ make_bundle [] empty_slurpy context_infos [thepatch]
  if wantfile
     then do fname <- get_output opts
             writeBinFile fname bundle
     else do thetargets <- get_targets opts
             promptChar "Do you really want to send this bug report? " "yn"
             sendEmail from (lt thetargets) patchname (get_cc opts)
                    $ make_email (unpackPS $ unlinesPS report) bundle "aaack"
             putStr $ "Successfully sent bug report to "++lt thetargets++".\n"
                 where lt [t] = t
                       lt [t,""] = t
                       lt (t:ts) = t++" , "++lt ts
                       lt [] = ""
reportbug_cmd _ _ = impossible

make_email :: String -> String -> String -> String
make_email contents bundle boundary =
    "DarcsURL: http://abridgegame.org/repos/darcs\n"++
    "Content-Type: multipart/mixed; boundary=\""++boundary++"\"\n"++
    "\n--"++boundary++"\n"++
    "Content-Type: text/plain\n\n"++
    contents++"\n"++
    "\n--"++boundary++"\n"++
    "Content-Type: text/x-darcs-patch\n"++
    "Content-Description: A darcs patch for your repository!\n\n"++
    bundle ++
    "\n--"++boundary++"\n"++
    "\n.\n\n"
\end{code}

Please, do try using darcs reportbug if you have a bug in darcs to report.
At the moment, this is still quite an experimental command.  It is sort of
a prototype for the idea of keeping bug reports within a darcs repository,
although it doesn't have any of the nice features one would want...

\begin{code}
decide_on_behavior :: [DarcsFlag] -> IO [DarcsFlag]
decide_on_behavior opts =
    if want_to opts then return opts
    else
      do wantfile <- wants_output opts
         if wantfile
          then return opts
          else return (Target "DarcsBug <droundy@abridgegame.org>":opts)
    where want_to [] = False
          want_to (Target _:_) = True
          want_to (_:os) = want_to os
\end{code}

\begin{code}
wants_output :: [DarcsFlag] -> IO Bool
wants_output (Output _:_) = return True
wants_output (_:flags) = wants_output flags
wants_output [] = return False
get_output :: [DarcsFlag] -> IO String
get_output (Output a:_) = return a
get_output (_:flags) = get_output flags
get_output [] = bug "in ReportBug: called get_output when wants_output is false."
\end{code}

\begin{code}
get_targets :: [DarcsFlag] -> IO [String]
get_targets flags =
    case catMaybes $ map towhom flags of
    [] -> do putStr "What is the target email address? "
             hFlush stdout
             liftM (:[]) getLine
    ts -> return $ ts
    where towhom (Target t) = Just t
          towhom _ = Nothing
\end{code}
